C $Header: /export/cvs/c34a1/source/moldyn/dlef.f,v 1.1.1.1 2006/08/17 02:29:56 brooks Exp $
C
C COPYRIGHT (C) 1991-1998 MOLDYN, INC.  ALL RIGHTS RESERVED
C ========================================================================
C
C THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF MOLDYN, INC.
C
C THIS SOFTWARE CONTAINS PROPRIETARY AND CONFIDENTIAL INFORMATION OF
C MOLDYN, INC., AND IS PROTECTED BY COPYRIGHT, TRADE SECRET AND OTHER 
C STATE AND FEDERAL LAWS.  USE, DISCLOSURE, REPRODUCTION OR DUPLICATION 
C IN WHOLE OR IN PART IS STRICTLY PROHIBITED WITHOUT THE PRIOR EXPRESS 
C WRITTEN CONSENT OF MOLDYN, INC. THE ABOVE COPYRIGHT NOTICE DOES NOT 
C EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION OF SUCH SOURCE CODE.
C
C                    SUBROUTINE - DLEF.F
C
C::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
c
      subroutine dlef(dmd,glm,ndmd,nlm,icbl,icrow,iccol)
c
c     DLEF code solves the sparse system of linear equations
c
c                      (L*L')*lambda = g                           
c
c     for position and velocity fitting in cmpic2.f after the matrix
c     D*M(-1)*D' has been factorized in dllt.f in the form L*L'
c     where L is an lower triangular matrix  
c
c     DLEF code is called by dlespa.f.
c
c     The inputs are:
c     dmd     1-dimensional array to store not-zero elements of 
c             L (represents the factorized D*M(-1)*D'), size (1 x ndmd).
c             Output from dllt.f.
c     glm     (g) Right-hand vector of the system of linear equations,
c             size (1 x nlm).
c     ndmd    Length of dmd (computed in dlespa.f).
c     nlm     Size of the full matrix D*M(-1)*D'.
c             In the case of position fitting: nlm = nconn,
c             i.e. is equal to the number of all constraints,
c             In the case of velocity fitting: nlm = nconn + 6.
c     icbl    Matrix (4 x nlm) computed in dlespa.f and describing
c             the sparsity of L for fast element-by-element
c             operations (Note:  factorization L*L' retains most of 
c             the structure of sparsity).
c             1st row:  Number of constraints in i-th row's
c                       constraint-block, i.e. number of constraints j
c                       interacting with each i-th constraint (producing
c                       not-zero elements), (j >= i).
c                       The diagonal elements are included.
c             2nd row:  # of first element of i-th row's constraint-block
c                       in 1-dimensional array (see icrow).
c             3rd row:  Number of constraints in j-th column's
c                       constraint-block, i.e. number of constrains i
c                       interacting with each j-th constraint (producing
c                       not-zero elements), (j > i). 
c             4th row:  # of first element of j-th column's constraint-block
c                       in 1-dimensional array (see iccol).
c 
c     icrow   1-dimensional array to store #s of interacting constraints j
c             for each i-th row's constraint-block (j >= i), size (1 x ndmd).
c             Computed in dlespa.f.  Consistent with arrangement of dmd.
c     iccol   1-dimensional array to store #s of interacting constraints i
c             for each j-th column's constraint-block (j >= i),
c             size (1 x ndmd).  Computed in dlespa.f.
c             Note:  Array iccol and two last rows of icbl can be restored
c                    from icrow and two first rows of icbl.  But they are
c                    precalculated to avoid extra cycling for very
c                    sparse matrix L (factorized D*M(-1)*D'). 
c
c     The outputs are:
c     glm     (lambda) Solution to the sparse system of linear equations.
c
      IMPLICIT NONE
c
c     Declare variables:
c
      integer ndmd,nlm                      ! Inputs
      double precision dmd(ndmd),glm(nlm)   ! Inputs/outputs
      double precision dmdi
      integer icbl(4,nlm)                   ! Inputs
      integer icrow(ndmd),iccol(ndmd)       ! Inputs
      integer i,j,k,ki,kk,ik,jc,kir,kic
      integer iintcr,iintcc,irw
      CHARACTER*128  RCSID
      DATA           RCSID
     &/'$RCSfile: dlef.f,v $$Revision: 1.1.1.1 $'/
c
c     Perform forward recursion
c     Initialize the forward recursion
      glm(1) = glm(1)/dmd(1)
c
c     Pass through each constrains' row (forward)
c     Note:  All calculations here are performed with the transpored matrix L'
c            but terms 'row' and 'column' are related to matrix L
c
      do i = 2, nlm
         iintcr = icbl(1,i)     !  Number of not-zero elements in i-th row
         kir = icbl(2,i) - 1    !  Pointer(-1) for i-th row's c-block
         iintcc = icbl(3,i) - 1 !  Number of not-zero elements in i-th column
c                               !  (diagonal elements are not included)
         kic = icbl(4,i) - 1    !  Pointer(-1) for i-th column's c-block
c
c        Pass through each not-zero element in the i-th column
c        (exept the last, diagonal, one)
         dmdi = glm(i)
         do k = 1, iintcc
            ik = iccol(kic+k)      !  Row's # (ik)
            irw = icbl(1,ik)       !  Length of 'ik'-th c-block
            ki = icbl(2,ik) - 1    !  Pointer(-1) for 'ik'-th c-block
c                                  !  in icrow (dmd)
c           Pass through each not-zero element of the 'ik'-th row
            do kk = 1, irw
c              Find element of 'ik'-th constraint-block
c              corresponding to i-th column
               if (icrow(ki+kk) .eq. i) then
c                  j = icrow(ki+kk)   
                  dmdi = dmdi - dmd(ki+kk)*glm(ik)
               end if
            end do
         end do   ! k
         glm(i) = dmdi/dmd(kir+1)
      end do   ! i
c
c     Perform backward recursion
c     Initialize the backward recursion
      glm(nlm) = glm(nlm)/dmd(ndmd)
c
c     Pass through each constrains' row (backward)
      do i = nlm-1, 1, -1
         iintcr = icbl(1,i)     !  Number of not-zero elements in i-th row
         kir = icbl(2,i) - 1    !  Pointer(-1) for i-th row's c-block
c
c        Pass through each not-zero element in the row (exept the first one)
         dmdi = glm(i)
         do jc = 2, iintcr
            j = icrow(kir+jc)
            dmdi = dmdi - dmd(kir+jc)*glm(j)
         end do   ! jc
         glm(i) = dmdi/dmd(kir+1)
      end do   ! i
c
      return 
      end








