C $Header: /export/cvs/c34a1/source/moldyn/dllt.f,v 1.1.1.1 2006/08/17 02:29:56 brooks Exp $
C
C COPYRIGHT (C) 1991-1998 MOLDYN, INC.  ALL RIGHTS RESERVED
C ========================================================================
C
C THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF MOLDYN, INC.
C
C THIS SOFTWARE CONTAINS PROPRIETARY AND CONFIDENTIAL INFORMATION OF
C MOLDYN, INC., AND IS PROTECTED BY COPYRIGHT, TRADE SECRET AND OTHER 
C STATE AND FEDERAL LAWS.  USE, DISCLOSURE, REPRODUCTION OR DUPLICATION 
C IN WHOLE OR IN PART IS STRICTLY PROHIBITED WITHOUT THE PRIOR EXPRESS 
C WRITTEN CONSENT OF MOLDYN, INC. THE ABOVE COPYRIGHT NOTICE DOES NOT 
C EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION OF SUCH SOURCE CODE.
C
C                    SUBROUTINE - DLLT.F
C
C::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
c
      subroutine dllt(dmd,ndmd,nlm,icbl,icrow,iccol,inform)
c
c     DLLT code performs factorization of the sparse positive-defined
c     symmetric matrix 
c
c                    D*M(-1)*D'  --->  L*L'
c
c     where L is an lower triangular matrix
c
c     Factorization is performed for solving the system of sparse LE
c
c                     [D*M(-1)*D']*lambda = g
c
c     The structure of sparsity corresponds to the problem of 
c     position and velocity fitting in cmpic2.f.
c     Factorization L*L' retains the structure of sparsity in the case
c     of tree topology (witt consequent numeration of bodies),
c     i.e. L is as sparsed as the lower triangular of D*M(-1)*D'.
c     In the case of closed-loop topology the matrix L has additional
c     not-zero elements (due to "twists" in topology).  The previous
c     version of dllt.f predicted those elements and provided exact
c     matrix L, but it was a quite complex algorithm.
c     This version of dllt.f provides matrix L assuming tree topology
c     (which means that matrix L will not be "exact" in the case of
c     closed-loops).  In this case the solution of LE obtained from L
c     (see dlef.f) is used as effective initial guess for iterative 
c     algorithm (see dloop.f).
c
c     DLLT code is called by dlespa.f.
c
c     The inputs are:
c     dmd     1-dimensional array to store not-zero elements of D*M(-1)*D'.
c             Takes into account the matrix's symmetry.
c             Arranged in dlespa.f into constraint-blocks for each row.
c             First constraint-block contains all not-zero elements
c             in the first row of U = L' left to the diagonal element
c             (starting from the diagonal element).  The same for the
c             second constraint-block, etc.
c     ndmd    Length of dmd (computed in dlespa.f).
c     nlm     Size of the full matrix D*M(-1)*D'.
c             In the case of position fitting: nlm = nconn,
c             i.e.number of all constraints,
c             In the case of velocity fitting: nlm = nconn + 6.
c     icbl    Matrix (4 x nlm) computed in dlespa.f and describing
c             the sparsity of L for fast element-by-element
c             operations.
c             1st row:  Number of constraints in i-th row's
c                       constraint-block, i.e. number of constraints j
c                       interacting with each i-th constraint (producing
c                       not-zero elements), (j >= i).
c                       The diagonal elements are included.
c             2nd row:  # of first element of i-th row's constraint-block
c                       in 1-dimensional array (see icrow).
c             3rd row:  Number of constraints in j-th column's
c                       constraint-block, i.e. number of constrains i
c                       interacting with each j-th constraint (producing
c                       not-zero elements), (j > i). 
c             4th row:  # of first element of j-th column's constraint-block
c                       in 1-dimensional array (see iccol).
c 
c     icrow   1-dimensional array to store #s of interacting constraints j
c             for each i-th row's constraint-block (j >= i), size (1 x ndmd).
c             Computed in dlespa.f.  Consistent with arrangement of dmd.
c     iccol   1-dimensional array to store #s of interacting constrains i
c             for each j-th  column's constraint-block (j >= i),
c             size (1 x ndmd).  Computed in dlespa.f. 
c             Note:  Array iccol and two last rows of icbl can be restored
c                    from icrow and two first rows of icbl.  But they are
c                    precalculated to avoid extra cycling for very
c                    sparse matrix L. 
c     inform  = 0 if factorization has been performed normally,
c             = 1 if factorization fails (e.g. dmd is not symmetric and
c             positive-definite).      
c
c     The outputs are:
c     dmd     Factorized D*M(-1)*D' (L) stored in 1-dimensional array
c             with the same structure as in input's dmd.  
c
      IMPLICIT NONE
c
c     Declare variables:
c
      integer ndmd,nlm,inform               ! Inputs
      double precision dmd(ndmd)            ! Inputs/outputs
      double precision dmdi
      integer icbl(4,nlm)                   ! Inputs
      integer icrow(ndmd),iccol(ndmd)       ! Inputs
      integer i,j,k,ki,kk,ik,jc,kir,kic
      integer iintcr,iintcc,irw,ind1,ind2
      CHARACTER*128  RCSID
      DATA           RCSID
     &/'$RCSfile: dllt.f,v $$Revision: 1.1.1.1 $'/
c
      inform = 0
c     Define 1st constraint-block of the factorized dmd
      if (dmd(1) .le. 0) then
         inform = 1
      end if
      dmdi = dsqrt(dmd(1))
      dmd(1) = dmdi
      iintcr = icbl(1,1)
      do j = 2, iintcr
         dmd(j) = dmd(j)/dmdi
      end do
c
c     Define 2nd, etc. constraint-blocks of the factorized dmd
c     Note:  Interacting constraints only with j >= i are stored in dmd
c
c     Pass through each i-th (i > 1) row
      do i = 2, nlm
         iintcr = icbl(1,i)     ! Number of not-zero elements in i-th row
         kir = icbl(2,i) - 1    ! Pointer(-1) for i-th row's c-block
         iintcc = icbl(3,i) - 1 ! Number of not-zero elements in i-th column
c                               ! (diagonal elements are not included)
         kic = icbl(4,i) - 1    ! Pointer(-1) for i-th column's c-block 
c
c        Pass through each not-zero element of the i-th row
         do jc = 1, iintcr
            j = icrow(kir+jc)
            dmdi = dmd(kir+jc)
c         
c           Make summation only for interacting constraints
c           (i-th (j=i) column is leading) 
            do k = 1, iintcc
                ik = iccol(kic+k)       ! Row's # (ik)
                irw = icbl(1,ik)        ! Length of 'ik'-th c-block
                ki = icbl(2,ik) - 1     ! Pointer(-1) for 'ik'-th c-block 
c                                       ! in icrow (dmd)
                ind1 = 0
                ind2 = 0
c
c               Pass through each not-zero element of the 'ik'-th row
                do kk = 1, irw    
c                  Find element of 'ik'-th constraint-block 
c                  corresponding to i-th column                
                   if (icrow(ki+kk) .eq. i) then
                      ind1 = kk
                   end if
c                  Find element of 'ik'-th constraint-block
c                  corresponding to j-th column
                   if (icrow(ki+kk) .eq. j) then
                      ind2 = kk
                   end if
                end do
c               Compute contribution from not-zero elements
                if (ind1.ne.0 .and. ind2.ne.0) then      
                   dmdi = dmdi - dmd(ki+ind1)*dmd(ki+ind2)
                end if
            end do
c
c           Compute i-th diagonal element of L
            if (i .eq. j) then
               if (dmdi .lt. 0) then
                  inform = 1
               end if
               dmd(kir+1) = dsqrt(dmdi)
c
c           Compute off-diagonal elements in the i-th row of
c           L (with j > i)
            else 
               dmd(kir+jc) = dmdi/dmd(kir+1)
            end if
         end do   ! jc
      end do   ! i
c
      return 
      end





