C $Header: /export/cvs/c34a1/source/moldyn/dloop.f,v 1.1.1.1 2006/08/17 02:29:56 brooks Exp $
C
C COPYRIGHT (C) 1991-1998 MOLDYN, INC.  ALL RIGHTS RESERVED
C ========================================================================
C
C THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF MOLDYN, INC.
C
C THIS SOFTWARE CONTAINS PROPRIETARY AND CONFIDENTIAL INFORMATION OF
C MOLDYN, INC., AND IS PROTECTED BY COPYRIGHT, TRADE SECRET AND OTHER 
C STATE AND FEDERAL LAWS.  USE, DISCLOSURE, REPRODUCTION OR DUPLICATION 
C IN WHOLE OR IN PART IS STRICTLY PROHIBITED WITHOUT THE PRIOR EXPRESS 
C WRITTEN CONSENT OF MOLDYN, INC. THE ABOVE COPYRIGHT NOTICE DOES NOT 
C EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION OF SUCH SOURCE CODE.
C
C                    SUBROUTINE - DLOOP.F
C
C::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
c
      subroutine dloop(dmd,gv,glm,ndmd,nlm,icbl,icrow,iccol)
c
c     DLOOP code realizes iterative Gauss-Seidel algorithm 
c     to solve the system of sparse LE
c
c                    [D*M(-1)*D']*lambda = g
c
c     Generally, an arbitrary initial guess for lambda can be used,
c     since the Gauss-Seidel is convergent due to the fact that the 
c     matrix [D*M(-1)*D'] is positive definite.
c
c     However, dloop.f works after dllt.f/dlef.f (which solve
c     the system of LE with the tree topology).  In the case of 
c     the closed-loop topology dloop.f will correct the "tree" solution. 
c
c     DLOOP code is called by dlespa.f.
c
c     The inputs are:
c     dmd     1-dimensional array to store not-zero elements of D*M(-1)*D'.
c             Takes into account the matrix's symmetry.
c             Arranged in dlespa.f into constraint-blocks for each row.
c             First constraint-block contains all not-zero elements
c             in the first row of U = L' left to the diagonal element
c             (starting from the diagonal element).  The same for the
c             second constraint-block, etc.
c     gv      (g) Right-hand vector of the system of linear equations,
c             size (1 x nlm).
c     glm     Initial guess for lambda, size (1 x nlm).
c     ndmd    Length of dmd (computed in dlespa.f).
c     nlm     Size of the full matrix D*M(-1)*D'.
c             In the case of position fitting: nlm = nconn,
c             i.e.number of all constraints,
c             In the case of velocity fitting: nlm = nconn + 6.
c     icbl    Matrix (4 x nlm) computed in dlespa.f and describing
c             the sparsity of L for fast element-by-element
c             operations.
c             1st row:  Number of constraints in i-th row's
c                       constraint-block, i.e. number of constraints j
c                       interacting with each i-th constraint (producing
c                       not-zero elements), (j >= i).
c                       The diagonal elements are included.
c             2nd row:  # of first element of i-th row's constraint-block
c                       in 1-dimensional array (see icrow).
c             3rd row:  Number of constraints in j-th column's
c                       constraint-block, i.e. number of constrains i
c                       interacting with each j-th constraint (producing
c                       not-zero elements), (j > i). 
c             4th row:  # of first element of j-th column's constraint-block
c                       in 1-dimensional array (see iccol).
c 
c     icrow   1-dimensional array to store #s of interacting constraints j
c             for each i-th row's constraint-block (j >= i), size (1 x ndmd).
c             Computed in dlespa.f.  Consistent with arrangement of dmd.
c     iccol   1-dimensional array to store #s of interacting constrains i
c             for each j-th  column's constraint-block (j >= i),
c             size (1 x ndmd).  Computed in dlespa.f. 
c             Note:  Array iccol and two last rows of icbl can be restored
c                    from icrow and two first rows of icbl.  But they are
c                    precalculated to avoid extra cycling for very
c                    sparse matrix L. 
c
c     The outputs are:
c     glm     (lambda) Solution to the system of LE.      
c
      IMPLICIT NONE
c
c     Declare variables:
c
      integer ndmd,nlm                      ! Inputs
      double precision dmd(ndmd),gv(ndmd)   ! Inputs
      double precision glm(ndmd)            ! Output        
      double precision dmdi,gvi
      integer icbl(4,nlm)                   ! Inputs
      integer icrow(ndmd),iccol(ndmd)       ! Inputs
      integer i,j,k,ki,kk,ik,jc,kir,kic
      integer iintcr,iintcc,irw,ind1,ind2
      integer iteri,itmax,icnt
      CHARACTER*128  RCSID
      DATA           RCSID
     &/'$RCSfile: dloop.f,v $$Revision: 1.1.1.1 $'/
c
c     Restore A (lower triangular) from L 
c
c     Pass through each i-th row (backward)
      do i = nlm,1,-1
         iintcr = icbl(1,i)     ! Number of not-zero elements in i-th row
         kir = icbl(2,i) - 1    ! Pointer(-1) for i-th row's c-block
         iintcc = icbl(3,i) - 1 ! Number of not-zero elements in i-th column
c                               ! (diagonal elements are not included)
         kic = icbl(4,i) - 1    ! Pointer(-1) for i-th column's c-block 
c
c        Pass through each not-zero element of the i-th row
c        (from right to left)
         do jc = iintcr,1,-1
            j = icrow(kir+jc)
            dmdi = dmd(kir+1)*dmd(kir+jc)
c         
c           Make summation only for interacting constraints
c           (i-th (j=i) column is leading) 
            do k = 1, iintcc
                ik = iccol(kic+k)       ! Row's # (ik)
                irw = icbl(1,ik)        ! Length of 'ik'-th c-block
                ki = icbl(2,ik) - 1     ! Pointer(-1) for 'ik'-th c-block 
c                                       ! in icrow (dmd)
                ind1 = 0
                ind2 = 0
c
c               Pass through each not-zero element of the 'ik'-th row
                do kk = 1, irw    
c                  Find element of 'ik'-th constraint-block 
c                  corresponding to i-th column                
                   if (icrow(ki+kk) .eq. i) then
                      ind1 = kk
                   end if
c                  Find element of 'ik'-th constraint-block
c                  corresponding to j-th column
                   if (icrow(ki+kk) .eq. j) then
                      ind2 = kk
                   end if
                end do
c               Compute contribution from not-zero elements
                if (ind1.ne.0 .and. ind2.ne.0) then      
                   dmdi = dmdi + dmd(ki+ind1)*dmd(ki+ind2)
                end if
            end do
            dmd(kir+jc) = dmdi
         end do   ! jc
      end do   ! i
c
c     Realize the Gauss-Seidel iterative algorithm
      iteri = 0
      itmax = 50
 100  continue
c
      icnt = 0
      do i = 1, nlm 
c        Perform multiplicaton [D*M(-1)*D']*glm but without diagonaly
         iintcr = icbl(1,i)     ! Number of not-zero elements in i-th row
         kir = icbl(2,i) - 1    ! Pointer(-1) for i-th row's c-block
         iintcc = icbl(3,i) - 1 ! Number of not-zero elements in i-th column
c                               ! (diagonal elements are not included)
         kic = icbl(4,i) - 1    ! Pointer(-1) for i-th column's c-block 
         gvi = 0.0d0
c
c        Make summation with respect to not-zero elements in the row
c        (right to the diagonal)
         do jc = 2, iintcr 
            gvi = gvi + dmd(kir+jc)*glm(icrow(kir+jc))
         end do
c        Make summation with respect to not-zero elements in the row
c        (left to the diagonal, or above the diagonal in the column)
         do k = 1, iintcc 
            ik = iccol(kic+k)     ! Row's # (ik)
            irw = icbl(1,ik)        ! Length of 'ik'-th c-block
            ki = icbl(2,ik) - 1   ! Pointer(-1) for 'ik'-th c-block 
c                                 ! in icrow (dmd)
            ind1 = 0
c           Pass through each not-zero element of the 'ik'-th row
            do kk = 1, irw    
c              Find element of 'ik'-th constraint-block 
c              corresponding to i-th column                
               if (icrow(ki+kk) .eq. i) then
                  ind1 = kk
               end if
            end do
            gvi = gvi + dmd(ki+ind1)*glm(ik)
         end do
c
c        Compute new glm (Gauss-Seidel update)
         glm(i) = (gv(i) - gvi)/dmd(kir+1)
c        Compute "new" right-hand part (to compare)
         gvi = gvi + dmd(kir+1)*gv(i)
c
c        Check if the i-th row is satisfied
         if (dabs(gvi-gv(i)).lt.1.0d-10) then
            icnt = icnt + 1       
         end if 
      end do
c
c     Check the stop criteria for the iterative algorithm
      if (icnt.eq.nlm .or. iteri.eq.itmax) then
         go to 200
      end if
      iteri = iteri + 1
      go to 100
c
 200  continue
      return 
      end





