c $Header: /export/cvs/c34a1/source/moldyn/mtsratio.f,v 1.1.1.1 2006/08/17 02:29:56 brooks Exp $
c
C COPYRIGHT (C) 1991-1995 MOLDYN, INC.  ALL RIGHTS RESERVED
C ========================================================================
C
C THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF MOLDYN, INC.
C
C THIS SOFTWARE CONTAINS PROPRIETARY AND CONFIDENTIAL INFORMATION OF
C MOLDYN, INC., AND IS PROTECTED BY COPYRIGHT, TRADE SECRET AND OTHER 
C STATE AND FEDERAL LAWS.  USE, DISCLOSURE, REPRODUCTION OR DUPLICATION 
C IN WHOLE OR IN PART IS STRICTLY PROHIBITED WITHOUT THE PRIOR EXPRESS 
C WRITTEN CONSENT OF MOLDYN, INC. THE ABOVE COPYRIGHT NOTICE DOES NOT 
C EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION OF SUCH SOURCE CODE.
C
C
      SUBROUTINE mtsratio(dt,calibr,dt_max,
     &                    NumBin,BinCount,BinSize,BinStart,
     &                    nstages,mratio,lcmmts,speedup)
C
C     PURPOSE:
C     This subroutine finds an optimal MTS architecture 
C     (number of MTS stages and MTS ratios) given the histogram
C     of relative time scales for non-bonded interactions and the calibration
C     factor which puts the relative time scales on the absolute time axis.
C     The time steps for the MTS stages are defined by multiplying 
C     the fastest time scale (set as an integration step TIMESTEP 
C     in the CHARMM/MBO(N)D input file) by the MTS ratios.
C
C     The optimality criterion for MTS architecture 
C     is the speed-up factor of MTSL over the regular Lobatto integrator
C     (with the fastest time step defined by TIMESTEP).
C
C     The subroutine performs a systematical search over all possible 
C     MTS architectures from the given set of candidates sequentially 
C     considering MTS with larger and larger number of stages (bins).
C     The optimization process is stopped after the algorithm detects that
C     adding a new MTS stage does not improve criterion more than the
C     specified tolerance.  Also, at each MTS stage, the subroutine assumes
C     that some small portion of the speed-up criterion may be traded-off for 
C     a better LCM (Least Common Multiple) of MTS ratios.  The rules 
C     of trade-off are specified in the corresponding tolerance parameters.
C
C     The subroutine is called only once in the beginning 
C     of the integration process assuming that the initial histogram
C     for relative time scales does not change significantly 
C     (due to the fact that a set of interacting pairs within the specified
C     cutoff is approximately the same). 
C
C
      IMPLICIT NONE
#include "PARAM.FOR"
#include "NUMBRS.INC"

      CHARACTER*128  RCSID
      DATA           RCSID
     &/'$RCSfile: mtsratio.f,v $$Revision: 1.1.1.1 $'/
c
c     The inputs are:
c
      INTEGER          NumBin     !  The number of bins in the histogram
                                  !  for relative time scales.
      INTEGER          BinCount(NumBin)
                                  !  The number of interaction pairs in
                                  !  each bin of the histogram
                                  !  Note:  BinCount will be turned into 
                                  !         "integrated" BinCount. 
      DOUBLE PRECISION dt         !  Integration time step (the fastest).
      DOUBLE PRECISION calibr     !  Calibration factor to multiply time 
                                  !  scales.
      DOUBLE PRECISION dt_max     !  Largest permitted timestep
      DOUBLE PRECISION BinSize    !  Size of each bin of the histogram.
      DOUBLE PRECISION BinStart   !  Leftmost edge of the first (lowest) bin.
c
c     The outputs are:
      INTEGER          nstages    !  The optimal number of MTS stages.
      INTEGER          mratio(maxstages)
                                  !  The optimal MTS ratios.
                                  !  (including the ratio for the first stage
                                  !  which is normally equal to 1).
      INTEGER          lcmmts     !  LCM (Least Common Multiple)
                                  !  for optimal MTS ratios.
      DOUBLE PRECISION speedup    !  Optimal speed-up factor 
                                  !  (related to the percentage of FF, perff).
c
c     Local variables:
c
      LOGICAL          NewStage   !  Indicator to open a new MTS stage.
      INTEGER          max_ratio  !  Maximum value of the MTS ratio.
      INTEGER          i          !  DO loop index.
      INTEGER          j          !  DO loop index.
      INTEGER          ind(maxstages)
                                  !  Indices to indicate the current 
                                  !  MTS ratio for each MTS stage.
      INTEGER          ind_init(maxstages)
                                  !  Indices to indicate the initial 
                                  !  MTS ratio for each MTS stage
                                  !  (changes inside nested loops).
      INTEGER          ind_ceil(maxstages)
                                  !  Upper limits for indices ind.
      INTEGER          nbin       !  The current n in n-bin MTS.
      INTEGER          mratio_cur(maxstages)
                                  !  MTS ratios for the current MTS.
      INTEGER          kratios(100)
                                  !  List of possible MTS ratios
                                  !  (not necessarily 1,2,3,4,5,6  etc.)
      INTEGER          lcm        !  Function name to compute LCM.
      INTEGER          lcmmts_cur !  Current LCM for MTS ratios.
      INTEGER          lcm_max    !  Maximal LCM (cutoff).
      INTEGER          nbin1      !  nbin - 1.
      INTEGER          BinSum     !  For summation of bin counts.
      DOUBLE PRECISION perff      !  Percentage of the force-field (FF)
                                  !  evaluations in the total CPU time.
                                  !  Note:  perff is needed only for  
                                  !  assessment of the absolute value of the 
                                  !  speed-up factor (ranking of MTS 
                                  !  architectures does not depend on perff).
      DOUBLE PRECISION scale_max  !  Maximum time scale (calibrated). 
      DOUBLE PRECISION BinSize_c  !  Calibrated BinSize.
      DOUBLE PRECISION BinStart_c !  Calibrated BinStart.
      DOUBLE PRECISION f_speedup  !  Function name for speed-up factor.
      DOUBLE PRECISION speedup_cur
                                  !  Current speed-up factor. 
      DOUBLE PRECISION speed_pr_st    
                                  !  Optimal speedup at previous stage.
      DOUBLE PRECISION tol_st     !  Relative tolerance in minimizing 
                                  !  speed-up factor (between MTS stages).
      DOUBLE PRECISION tol_lcm    !  Relative tolerance in minimizing
                                  !  speed-up factor (to choose a better LCM).
      DOUBLE PRECISION sprel      !  Relative increase in speedup.
      DOUBLE PRECISION dt_edge, dt_edge_max
                                  !  Edge for time step
      DOUBLE PRECISION t1,t2,time,time_max
                                  !  CPU timing.
c
c     Set control parameters
      time_max = 12.0D1    !  Max CPU time (sec)
      perff = 0.7D0       !  FF percentage in CPU time
      tol_st = 0.03D0     !  Tolerance for speed-up (between stages)
      tol_lcm = 0.02D0    !  Tolerance for speed-up (LCM)

      dt_edge_max = dt_max      !  Maximal edge for time step (ps)

c
c     Start the CPU timer
      CALL msecond(t1)
c
c     Truncate the zero right part of the histogram 
c     and set the new number of histogram's bins
      DO i = NumBin, 1, -1
         IF (BinCount(i) .NE. 0) GO TO 21
      END DO
 21   NumBin = i
c
c     Turn the "density" histogram into the "integrated" histogram
c     by consequtive integrating over histogram bins
      BinSum = 0
      DO i = 1, NumBin
         BinSum = BinSum + BinCount(i)
         BinCount(i) = BinSum
      END DO
c
c     Calibrate the relative time scales
c     (place the histogram on absolute time axis)
      BinStart_c = calibr*BinStart
      BinSize_c = calibr*BinSize
c
c     Select the maximum value of the MTS ratio
      scale_max = BinStart_c + NumBin*BinSize_c
      max_ratio = int(scale_max/dt)
      dt_edge = max_ratio*dt
      IF (dt_edge .GT. dt_edge_max) THEN
         max_ratio = int(dt_edge_max/dt)
      END IF
c
c     Set the cutoff value for LCM of MTS ratios (based on max_ratio)
      lcm_max = 2*max_ratio
c
c     Set possible MTS ratios 
c     Note:  The current 1,2,3,4,5,6 set-up may be changed
      IF (max_ratio .GT. 100) max_ratio = 100
      DO i = 1, max_ratio
         kratios(i) = i
      END DO
c
c     Cycle through all possible MTS architectures 
c     starting from 2-bin MTS and going to n-bin MTS until 
c     the speed-up factor is not improved more than the specified tolerance
c
      nbin = 2
      mratio_cur(1) = 1
      ind(1) = 1
      speedup = ONE
      speed_pr_st = ONE
      NewStage = .FALSE.
c
 1    CONTINUE
c
c     Start a new MTS stage 
c
c     Initialize the indices and their ceiling values for a new MTS stage
      DO i = 2, nbin
         ind_init(i) = i
         ind(i) = i
         ind_ceil(i) = max_ratio - (nbin - i)
      END DO
c
 3    CONTINUE
c
c     Check the CPU timer
      CALL msecond(t2)
      time = t2 - t1
      IF (time .GT. time_max) GO TO 100   !  The CPU time exceeded the limit
c
c     Set or re-set MTS architecture within the current MTS stage
      DO i = 2, nbin
         mratio_cur(i) = kratios(ind(i))
      END DO
c
c     The following algorithm manages the change of indices for MTS stages. 
c     When the algorithm detects that the index for higher MTS stage 
c     reaches its ceiling value, the algorithm initializes this value again.
c     When all indices of the MTS stages are "worked out" the algorithm
c     signals that a new stage should be open
c
      DO i = nbin, 2, -1   ! Start from the highest bin
         IF (ind(i) .LT. ind_ceil(i)) THEN
            ind(i) = ind(i) + 1   ! Increase the index of MTS ratio
            GO TO 4
         ELSE
            IF (i .GT. 2) THEN
               DO j = nbin, i, -1
                  ind_init(j) = ind_init(i) + 1 + j - i
                  ind(j) = ind_init(j)
                         ! Initialize again the index of MTS ratio
                         ! Note:  Index for the 2nd MTS bin
                         !        (the 1st changeable stage)
                         !        does not need to be initialized
               END DO
            ELSE
               NewStage = .TRUE.   ! Open a new MTS stage
            END IF
         END IF
      END DO
c
 4    CONTINUE
c
c     Compute the LCM for MTS stages
      nbin1 = nbin - 1
      lcmmts_cur = lcm(nbin1,mratio_cur(2))
c
      IF (lcmmts_cur .LE. lcm_max) THEN
c
c     Compute the speed-up factor for the current MTS architecture
         speedup_cur = f_speedup(nbin,mratio_cur,dt,perff,
     &                 NumBin,BinCount,BinSize_c,BinStart_c)
c
c     Memorize the best MTS architecture at this moment 
c     Note:   It is possible to trade some speed-up factor (within tolerance)
c             for a better LCM
         sprel = (speedup_cur - speedup)/speedup
         IF (speedup_cur .GT. speedup) THEN     
                                ! **** Make dependent on sprel and tolerance
            speedup = speedup_cur
            nstages = nbin
            lcmmts = lcmmts_cur
            DO i = 1, nbin
               mratio(i) = mratio_cur(i)
            END DO
         END IF
c     
      END IF
c
      IF (NewStage) THEN
c     Go to next-level MTS (add one more bin)
         nbin = nbin + 1
         IF (nbin .GT. maxstages) GO TO 100  ! Maximum number of bins 
                                             ! has been exceeded
         sprel = (speedup - speed_pr_st)/speed_pr_st
         IF (sprel .LT. tol_st) GO TO 100    ! The speed-up factor 
                                             ! reached saturation at this stage
         NewStage = .FALSE.
         speed_pr_st = speedup
         GO TO 1   ! Go to a new MTS stage
      ELSE
         GO TO 3
      END IF
c
 100  CONTINUE
c
      RETURN
      END
c
c:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
c
      FUNCTION f_speedup(nstages,mratio,dt,perff,
     &                 NumBin,BinCount,BinSize_c,BinStart_c)
C
C     PURPOSE:
C     This function computes the speed-up of the MTSL over the 
C     regular Lobatto integrator given the architecture of the MTS
C     (the number of MTS stages and the MTS ratios). 
C
      IMPLICIT NONE
#include "PARAM.FOR"
#include "NUMBRS.INC"
c 
c     The inputs are:
c
      INTEGER          nstages    !  The number of MTS stages.
      INTEGER          mratio(nstages)
                                  !  The MTS ratios.
                                  !  (including the ratio for the first stage
                                  !  which is normally equal to 1).
      INTEGER          NumBin     !  The number of bins in the histogram
                                  !  for relative time scales.
      INTEGER          BinCount(NumBin)
                                  !  "Integrated" histogramm BinCount
                                  !  (input in mtsratio.f). 
      DOUBLE PRECISION dt         !  Integration time step (the fastest).
      DOUBLE PRECISION perff      !  Percentage of the force-field (FF)
                                  !  evaluations in the total CPU time.
                                  !  Note:  perff is needed only for  
                                  !  assessment of the absolute value of the 
                                  !  speed-up factor (ranking of MTS 
                                  !  architectures does not depend on perff).
      DOUBLE PRECISION BinSize_c  !  Size of each bin of the histogram
                                  !  (calibrated).
      DOUBLE PRECISION BinStart_c !  Leftmost edge of the first (lowest) bin
                                  !  (calibrated).
c
c     The outputs are:
      DOUBLE PRECISION f_speedup    !  The speed-up factor.
c
c     Local variables:
c
      INTEGER          i          !  DO loop index.
      INTEGER          ib_left    !  The left index edge of MTS stage.
      INTEGER          ib_right   !  The right index edge of MTS stage.
      DOUBLE PRECISION Sstage(maxstages)
                                  !  Fraction of intercating pairs for 
                                  !  each MTS stage (Stage's weight).
      DOUBLE PRECISION BinC_left  !  The bin count for the left edge
                                  !  of MTS stage.
      DOUBLE PRECISION BinC_right !  The bin count for the right edge
                                  !  of MTS stage.
      DOUBLE PRECISION sums       !  Result of summation.
c
c
c     Form vector of weights for each MTS stage
c     (the fraction of interactions at the stage)
c
      DO i = 1, nstages
c
c     Form the left edge for the MTS stage and compute the bin count 
         IF (i .NE. 1) THEN
            ib_left = int((dt*mratio(i) - BinStart_c)/BinSize_c)  
            IF (ib_left .LE. 0) THEN
               BinC_left = ZRO
            ELSEIF (ib_left .GT. NumBin) THEN
               BinC_left = BinCount(NumBin)
            ELSE
               BinC_left = BinCount(ib_left)
            ENDIF
         ELSE
            BinC_left = ZRO
         END IF
c
c     Form the right edge for the MTS stage and compute the bin count
         IF (i .NE. nstages) THEN
            ib_right = int((dt*mratio(i + 1) - BinStart_c)/BinSize_c)
            IF (ib_right .LE. 0) THEN
               BinC_right = ZRO
            ELSEIF (ib_right .GT. NumBin) THEN
               BinC_right = BinCount(NumBin)
            ELSE
               BinC_right = BinCount(ib_right) 
            ENDIF
         ELSE
            BinC_right = BinCount(NumBin)
         END IF
c
c     Compute the stage's weight 
         Sstage(i) = BinC_right - BinC_left
c
      END DO
c
c     Normalize Sstage
      sums = ZRO
      DO i = 1, nstages
         sums = sums + Sstage(i) 
      END DO
      DO i = 1, nstages
         Sstage(i) = Sstage(i)/sums
      END DO
c
c     Compute the speed-up factor
      sums = ZRO
      DO i = 1, nstages
         sums = sums + Sstage(i)/dble(mratio(i))
      END DO
      f_speedup = ONE/(ONE - perff*(ONE - sums))
c
      RETURN
      END
c
c:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
c
      FUNCTION lcm(n,num)
C
C     PURPOSE:
C     This function computes the least common multiple (LCM) 
C     for a series of integer numbers.
C
C     The inputs are:
      INTEGER n         !  Amount of integer numbers.
      INTEGER num(*)    !  Integer numbers.
C 
C     The output is :
      INTEGER lcm       !  The LCM.
C
C     Local variables:
      INTEGER i         !  DO loop index.
      INTEGER j         !  DO loop index.
      INTEGER multwo    !  Multiple of 2 numbers.
      INTEGER num1      !  First number from 2.
      INTEGER num2      !  Second number from 2.
c
c
c     Recursive algorithm for finding LCM
      num1 = num(1)
      DO i = 2, n
         num2 = num(i)
         DO j = 1, num2
            multwo = num1*j
            IF (mod(multwo,num1).EQ.0 
     &          .AND. mod(multwo,num2).EQ.0) THEN
               num1 = multwo
               GO TO 1
            END IF
         END DO
 1       CONTINUE
      END DO
c
      lcm = num1
c
      RETURN
      END









